// Copyright (C) 2018-2019  Nexedi SA and Contributors.
//                          Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

// wcfs_virtmem provides wcfs client integrated with user-spce virtmem layer.
//
// XXX provides isolated view.

// XXX wcfs_virtmem organization.

#include "wcfs_misc.h"
#include "wcfs.h"
#include "wcfs_watchlink.h"

#include <wendelin/bigfile/virtmem.h>
#include <wendelin/bigfile/ram.h>
//#include <wendelin/bug.h>

#include <golang/fmt.h>

#include <algorithm>
#include <string>
#include <vector>

#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>



using std::min;
using std::max;
using std::vector;

static string h(uint64_t v); // v -> 016x hex representation
#define h_(v)   (h(v).c_str())
static error mmap_zero_into_ro(void *addr, size_t size);
static tuple<uint8_t*, error> mmap_ro(os::File f, off_t offset, size_t size);
static error mmap_into_ro(void *addr, size_t size, os::File f, off_t offset);

// _File represent isolated file view under Conn.
//
// XXX doc      XXX naming -> _FileView ?
// XXX -> refptr ?
struct _File {
    Conn        wconn;
    zodb::Oid   foid;       // ZBigFile root object ID
    size_t      blksize;    // block size of this file  XXX -> off_t ?
    os::File    headf;      // file object of head/file
    off_t       headfsize;  // head/file size is known to be at least headfsize (size ↑=)

    dict<int64_t, zodb::Tid>  pinned; // {} blk -> rev   that wcfs already sent us for this file
    vector<_Mapping*>         mmaps;  // []_Mapping ↑blk_start    mappings of this file
};

// _Mapping represents one mapping of _File.
// XXX -> refptr ?
struct _Mapping {
    _File   *file;
    int64_t blk_start;  // offset of this mapping in file

    BigFileH *fileh;        // mmapped under this file handle
    uint8_t  *mem_start;    // mmapped memory [mem_start, mem_stop)
    uint8_t  *mem_stop;

    int64_t blk_stop() const {
        //XXX reenable
        //ASSERT((mem_stop - mem_start) % file->blksize == 0);
        return blk_start + (mem_stop - mem_start) / file->blksize;
    }

    error _remmapblk(int64_t blk, zodb::Tid at);
};


// connect creates new Conn viewing WCFS state as of @at.
pair<Conn, error> WCFS::connect(zodb::Tid at) {
    WCFS *wc = this;
    // XXX err ctx

    // TODO support !isolated mode

    WatchLink wlink;
    error err;
    tie(wlink, err) = wc->_openwatch();
    if (err != nil)
        return make_pair(nil, err);

    Conn wconn = adoptref(new _Conn());
    wconn->_wc      = wc;
    wconn->at       = at;
    wconn->_wlink   = wlink;

    context::Context pinCtx;
    tie(pinCtx, wconn->_pinCancel) = context::with_cancel(context::background());
    wconn->_pinWG = sync::NewWorkGroup(pinCtx);
    wconn->_pinWG->go([wconn](context::Context ctx) -> error {
        return wconn->_pinner(ctx);
    });

    return make_pair(wconn, nil);
}

// close releases resources associated with wconn.
// XXX what happens to file mmappings?
error _Conn::close() {
    _Conn& wconn = *this;
    // XXX err ctx
    error err;

    err = wconn._wlink->close();
    if (err != nil)
        return err;
    wconn._pinCancel();
    err = wconn._pinWG->wait();
    if (err != context::canceled)  // canceled - ok
        return err;

    // close all files - both that have no mappings and that still have opened mappings.
    // XXX after file is closed mappings continue to survive, but we can no
    // longer maintain consistent view.
    wconn._filemu.lock();
    defer([&]() {
        wconn._filemu.unlock();
    });

    for (auto _ : wconn._filetab) {
        auto f = _.second;
        f->headf->close();    // XXX err
        f->headf = nil;

        // XXX stop watching f
    }

    wconn._filetab.clear();
    return nil;
}

// _pinner receives pin messages from wcfs and adjusts wconn mappings.
error _Conn::_pinner(context::Context ctx) {  // XXX error -> where?
    _Conn& wconn = *this;

    // XXX panic/exc -> log.CRITICAL
#if 0
    // if pinner fails, wcfs will kill us.
    // log pinner exception so the error is not hidden.
    // print to stderr as well as by default log does not print to there.
    def _():
        exc = sys.exc_info()[1]
        if exc in (None, context.canceled): # canceled = .close asks pinner to stop
            return
        log.critical('pinner failed:', exc_info=1)
        print('CRITICAL: pinner failed:', file=sys.stderr)
        traceback.print_exc(file=sys.stderr)
        print('\nCRITICAL: wcfs server will likely kill us soon.', file=sys.stderr)
    defer(_)
#endif

    PinReq req;
    error  err;

    while (1) {
        err = wconn._wlink->recvReq(ctx, &req);
        if (err != nil) {
            // XXX -> err, handle EOF, abort on *
            return err; // XXX ok? (EOF - when wcfs closes wlink)
        }

        // we received request to pin/unpin file block. handle it
        wconn._pin1(&req);
    }
}

// pin1 handles one pin request received from wcfs.
void _Conn::_pin1(PinReq *req) {
    _Conn& wconn = *this;

    // XXX return error?

    // XXX defer: reply either ack or nak on error
#if 0
    // reply either ack or nak on error
    def _():
        ack = "ack"
        exc = sys.exc_info()[1]
        if exc is not None:
            ack = "nak: %s" % exc
        #req.reply(ack)
        ctx = context.background()  # XXX ok?
        wconn._wlink.replyReq(ctx, req, ack)
    defer(_)
#endif

    _File *f;
    bool  ok;
    wconn._filemu.lock();
    tie(f, ok) = wconn._filetab.get(req->foid);
    if (!ok) {
        wconn._filemu.unlock();
        // XXX err = we are not watching the file - why wcfs sent us this update?
    }


    // XXX relock wconn -> f

    for (auto mmap : f->mmaps) {    // XXX use ↑blk_start for binary search
        if (!(mmap->blk_start <= req->blk && req->blk < mmap->blk_stop()))
            continue;   // blk ∉ mmap

        // XXX reenable
        //trace("\tremmapblk %d @%s" % (req->blk, (h(req.at) if req.at else "head")))
        // check if virtmem did not dirtied page corresponding to this block already
// XXX reenable (link to libbigfile dso)
#if 0
        virt_lock();
        TODO (mmap->file->blksize != mmap->fileh->ramh->ram->pagesize);
        if (!__fileh_page_isdirty(mmap->fileh, req->blk)) {
#endif
            mmap->_remmapblk(req->blk, req->at);    // XXX err
#if 0
    }
        virt_unlock();
#endif
        //trace("\t-> remmaped");   XXX
    }

    // update f.pinned
    // XXX do it before ^^^ remmapblk (so that e.g. concurrent
    // discard/writeout see correct f.pinned) ?
    if (req->at == TidHead) {
        f->pinned.erase(req->blk);      // unpin to @head
    }
    else {
        f->pinned[req->blk] = req->at;
    }

    wconn._filemu.unlock();
}

// mmap creates file mapping representing file[blk_start +blk_len) data as of wconn.at database state.
pair<_Mapping*, error> _Conn::mmap(zodb::Oid foid, int64_t blk_start, int64_t blk_len) {
    _Conn& wconn = *this;
    // XXX err ctx
    // XXX (blk_start + blk_len) * blk_size overflow
    error err;

    if (blk_start < 0)
        panic("blk_start < 0");
    if (blk_len < 0)
        panic("blk_len < 0");

    int64_t blk_stop = blk_start + blk_len;     // XXX overflow

    wconn._filemu.lock();
    defer([&]() {
        wconn._filemu.unlock();
    });

    _File* f; bool ok;
    tie(f, ok) = wconn._filetab.get(foid);
    if (f == nil) {
        f = new _File();    // XXX free f on error return (will be automatic if f = refptr)
        f->wconn    = newref(&wconn);    // XXX newref -> simpler?
        f->foid     = foid;
        tie(f->headf, err)
                    = wconn._wc->_open(fmt::sprintf("head/bigfile/%s", h_(foid)));
        if (err != nil)
            return make_pair(nil, err);
        //f.pinned    = {}
        //f.mmaps     = []

        struct stat st;
        err = f->headf->stat(&st);
        if (err != nil)
            return make_pair(nil, err);
        f->blksize   = st.st_blksize;
        f->headfsize = st.st_size;
        if (!(f->headfsize % f->blksize == 0))
            return make_pair(nil, fmt::errorf("wcfs bug: head/file size %% blksize != 0"));

        wconn._filetab[foid] = f;

        // start watching f
        string ack;
        tie(ack, err) = wconn._wlink->sendReq(context::background(), fmt::sprintf("watch %s @%s", h_(foid), h_(wconn.at)));
        if (err != nil)
            return make_pair(nil, err);
        if (ack != "ok") {
            // XXX unregister f from _filetab
            // XXX vvv -> errctx?
            return make_pair(nil, fmt::errorf("@%s: mmap f<%s>[%zd +%zd): %s", h_(wconn.at), h_(foid), blk_start, blk_len, ack.c_str()));
        }
    }

    // XXX relock wconn -> f ?

    // create memory with head/f mapping and applied pins
    // mmap-in zeros after f.size (else access to memory after file.size will raise SIGBUS)
    int64_t start = blk_start*f->blksize;
    uint8_t *mem_start, *mem_stop;
    tie(mem_start, err) = mmap_ro(f->headf, start, blk_len*f->blksize);
    if (err != nil)
        return make_pair(nil, err);
    mem_stop = mem_start + blk_len*f->blksize;

    int64_t stop = blk_stop*f->blksize;
    if (stop > f->headfsize) {
        uint8_t *zmem_start = mem_start + (max(f->headfsize/*XXX -1 ?*/, start) - start);
        err = mmap_zero_into_ro(zmem_start, mem_stop - zmem_start);
        if (err != nil)
            return make_pair(nil, err);
    }

    _Mapping* mmap = new _Mapping();
    mmap->file      = f;
    mmap->blk_start = blk_start;
    mmap->mem_start = mem_start;
    mmap->mem_stop  = mem_stop;

    for (auto _ : f->pinned) {  // XXX keep f.pinned ↑blk and use binary search?
        int64_t    blk = _.first;
        zodb::Tid  rev = _.second;
        if (!(blk_start <= blk && blk < blk_stop))
            continue;   // blk out of this mapping
        mmap->_remmapblk(blk, rev); // XXX err?
    }

    f->mmaps.push_back(mmap);    // XXX keep f.mmaps ↑blk_start

    return make_pair(mmap, nil);
}

// resync resyncs connection and its mappings onto different database view.
error _Conn::resync(zodb::Tid at) {
    _Conn& wconn = *this;
    // XXX err ctx

    // XXX locking

    for (auto fit : wconn._filetab) {
        zodb::Oid  foid = fit.first;
        _File      &f   = *fit.second;

        // XXX if file has no mappings and was not used during whole prev
        // cycle - forget and stop watching it

        // update f.headfsize and remmap to head/f zero regions that are now covered by head/f
        struct stat st;
        error err = f.headf->stat(&st);
        if (err != nil)
            return err;

        if ((size_t)st.st_blksize != f.blksize)     // blksize must not change
            return fmt::errorf("wcfs bug: blksize changed: %zd -> %ld", f.blksize, st.st_blksize);
        auto headfsize = st.st_size;
        if (!(f.headfsize <= headfsize))   // head/file size ↑=
            return fmt::errorf("wcfs bug: head/file size not ↑=");
        if (!(headfsize % f.blksize == 0))
            return fmt::errorf("wcfs bug: head/file size %% blksize != 0");

        for (auto mmap : f.mmaps) {
            printf("  resync -> %s: unzero [%lu:%lu)", h_(at), f.headfsize/f.blksize, headfsize/f.blksize);
            uint8_t *mem_unzero_start = min(mmap->mem_stop,
                            mmap->mem_start + (f.headfsize - mmap->blk_start*f.blksize));
            uint8_t *mem_unzero_stop  = min(mmap->mem_stop,
                            mmap->mem_start + (headfsize   - mmap->blk_start*f.blksize));

            if (mem_unzero_stop - mem_unzero_start > 0) {
                err = mmap_into_ro(mem_unzero_start, mem_unzero_stop-mem_unzero_start, f.headf, f.headfsize);
                if (err != nil)
                    return err;
            }
        }

        f.headfsize = headfsize;


        string ack;
        tie(ack, err) = wconn._wlink->sendReq(context::background(), fmt::sprintf("watch %s @%s", h_(foid), h_(at)));
        if (err != nil)
            return err;
        if (ack != "ok") {
            // XXX unregister f from _filetab
            // XXX vvv -> errctx?
            return fmt::errorf("resync @%s -> @%s: f<%s>: %s", h_(wconn.at), h_(at), h_(foid), ack.c_str());
        }
    }

    wconn.at = at;
    return nil;
}

// _remmapblk remmaps mapping memory for file[blk] to be viewing database as of @at state.
//
// at=TidHead means unpin to head/ .
// NOTE this does not check whether virtmem already mapped blk as RW.
error _Mapping::_remmapblk(int64_t blk, zodb::Tid at) {
    // XXX err context? blk #<blk> @<at>
    _Mapping *mmap = this;

    //XXX reenable
    //ASSERT(mmap->blk_start <= blk && blk < mmap->blk_stop());
    _File *f = mmap->file;
    error err;

    uint8_t *blkmem = mmap->mem_start + (blk - mmap->blk_start)*f->blksize;
    os::File fsfile;
    bool fclose = false;
    if (at == TidHead) {
        fsfile = f->headf;
    }
    else {
        // TODO share @rev fd until wconn is resynced?
        tie(fsfile, err) = f->wconn->_wc->_open(
                fmt::sprintf("@%s/bigfile/%s", h_(at), h_(f->foid)));
        if (err != nil)
            return err;
        fclose = true;
    }
    defer([&]() {
        if (fclose)
            fsfile->close();
    });

    struct stat st;
    err = fsfile->stat(&st);
    if (err != nil)
        return err;
    if ((size_t)st.st_blksize != f->blksize)
        return fmt::errorf("wcfs bug: blksize changed: %zd -> %ld", f->blksize, st.st_blksize);

    // block is beyond file size - mmap with zeros - else access to memory
    // after file.size will raise SIGBUS. (assumes head/f size ↑=)
    if ((blk+1)*f->blksize > (size_t)st.st_size) {
        err = mmap_zero_into_ro(blkmem, 1*f->blksize);
        if (err != nil)
            return err;
    }
    // block is inside file - mmap file data
    else {
        err = mmap_into_ro(blkmem, 1*f->blksize, fsfile, blk*f->blksize);
        if (err != nil)
            return err;
    }

    return nil;
}

// XXX _Mapping::remmap_blk

// XXX _Mapping::unmap


// ---- WCFS raw file access ----

// _path returns path for object on wcfs.
// - str:        wcfs root + obj;
string WCFS::_path(const string &obj) {
    WCFS *wc = this;

    return wc->mountpoint + "/" + obj;
}

tuple<os::File, error> WCFS::_open(const string &path, int flags) {
    WCFS *wc = this;
    string path_ = wc->_path(path);
    return os::open(path_, flags);
}


// ---- misc ----

static string h(uint64_t v) {
    return fmt::sprintf("%016lx", v);
}

// mmap_zero_into_ro mmaps read-only zeros into [addr +size) so that region as all zeros.
// created mapping, even after it is accessed, does not consume memory.
static error _mmap_zero_into_ro(void *addr, size_t size);
static error mmap_zero_into_ro(void *addr, size_t size) {
    error err = _mmap_zero_into_ro(addr, size);
    if (err != nil)
        err = fmt::errorf("mmap zero"); // XXX -> xerr::contextf?
    return err;
}
static error _mmap_zero_into_ro(void *addr, size_t size) {
    // mmap /dev/zero with MAP_NORESERVE and MAP_SHARED
    // this way the mapping will be able to be read, but no memory will be allocated to keep it.
    os::File z;
    error err;
    tie(z, err) = os::open("/dev/zero");
    if (err != nil)
        return err;
    defer([&]() {
        z->close();
    });
    err = mm::map_into(addr, size, PROT_READ, MAP_SHARED | MAP_NORESERVE, z, 0);
    if (err != nil)
        return err;
    return nil;
}

// mmap_ro mmaps read-only fd[offset +size).
// The mapping is created with MAP_SHARED.
static tuple<uint8_t*, error> mmap_ro(os::File f, off_t offset, size_t size) {
    return mm::map(PROT_READ, MAP_SHARED, f, offset, size);
}

// mmap_into_ro mmaps read-only fd[offset +size) into [addr +size).
// The mapping is created with MAP_SHARED.
static error mmap_into_ro(void *addr, size_t size, os::File f, off_t offset) {
    return mm::map_into(addr, size, PROT_READ, MAP_SHARED, f, offset);
}


_Conn::_Conn()  {}
_Conn::~_Conn() {}
void _Conn::decref() {
    if (__decref())
        delete this;
}
